/*	
 *	framz80.c
 *
 *	P/ECE Z80 Emulator
 *
 *	CLiP - Common Library for P/ECE
 *	Copyright (C) 2001-2003 Naoyuki Sawa
 *
 *	* Sat Nov 22 06:00:00 JST 2003 Naoyuki Sawa
 *	- 쐬JnB
 */
#include "clip.h"
#include "clipz8op.h"	/*Z80G~[^p}N*/

/****************************************************************************
 *
 ****************************************************************************/

/* Z80G~[Vs܂B
 * [in]
 *	z80		Z80\́B
 *	cycle		sTCNB
 * [note]
 *	sTCNcycleɒB邩A܂HALTŏI܂B
 */
#ifndef Z80_ASM	/************************************************************/

void
z80_run(Z80* z80, int cycle)
{
	unsigned char code;
	const Z80OP* op;
	unsigned char arg1;
	unsigned char arg2;

	arg1 = arg2 = 0; /* x} */
	CYCLE = cycle;
	while(!HALT && CYCLE > 0) {
#ifdef Z80_TRACE
		z80_dump(z80);
#endif /*Z80_TRACE*/
		Z80TRACE("%04x: ", z80->pc);
		FETCH(code);
		op = &z80_op_table[code];
		if(op->argc >= 1) {
			FETCH(arg1);
			if(op->argc >= 2) {
				FETCH(arg2);
			}
		}
		EXEC();
		Z80TRACE("\n");
	}
}

#else	/********************************************************************/

/* * Fri Nov 28 06:00:00 JST 2003 Naoyuki Sawa
 * - argc̏o䗦uargc=0argc=1argc=2vƉ肵č܂B
 *   z80op_CB/ED/DD/FD/DDCB/FDCBɂĂlłB
 */
asm("
	.code
	.align 1
	.global z80_run
z80_run:
	pushn %r3
	xsub %sp, %sp, 4
	;
	ld.w %r0, %r12			; z80
	xld.w %r3, [%r0+48]		; z80->read
	xld.w [%r0+36], %r13		; CYCLE
	xcmp %r13, 0			; if(CYCLE <= 0) break
	xjrle z80_run_L50
	;
z80_run_L10:
	xld.ub %r10, [%r0+35]		; if(HALT) break
	xcmp %r10, 0
	xjrne z80_run_L50
	;
	xld.uh %r1, [%r0+22]		; PC
	;
	ld.w %r12, %r0			; code = z80->read(z80, PC++)
	ld.uh %r13, %r1
	call.d %r3
	add %r1, 1			; *delay*
	ld.ub %r13, %r10
	;
	xld.w %r2, z80_op_table		; op = &z80_op_table[code]
	xsll %r10, 3
	add %r2, %r10
	;
	xld.ub %r10, [%r2+4]		; argc == 0 ?
	xcmp %r10, 0	
	xjreq z80_run_L20
	xld.b [%sp+0], %r13		; (save code)
	;
	ld.w %r12, %r0			; arg1 = z80->read(z80, PC++)
	ld.uh %r13, %r1
	call.d %r3
	add %r1, 1			; *delay*
	ld.ub %r14, %r10
	;
	xld.ub %r10, [%r2+4]		; argc == 1 ?
	xcmp %r10, 1	
	xjreq z80_run_L30
	xld.b [%sp+1], %r14		; (save arg1)
	;
	ld.w %r12, %r0			; arg2 = z80->read(z80, PC++)
	ld.w %r13, %r1
	call.d %r3
	add %r1, 1			; *delay*
	ld.ub %r15, %r10
	;
	xld.ub %r14, [%sp+1]		; (restore arg1)
z80_run_L30:
	xld.ub %r13, [%sp+0]		; (restore code)
z80_run_L20:
	;
	xld.w %r10, [%r2+0]		; op->proc(z80, code, arg1, arg2)
	xld.h [%r0+22], %r1		; PC *anti-interlock*
	call.d %r10
	ld.w %r12, %r0			; *delay*
	;
	xld.w %r10, [%r0+36]		; CYCLE -= op->cycle
	xld.ub %r11, [%r2+5]
	sub %r10, %r11
	xld.w [%r0+36], %r10
	xjrgt z80_run_L10		; if(CYCLE <= 0) break
z80_run_L50:
	;
	xadd %sp, %sp, 4
	popn %r3
	ret
");

#endif	/********************************************************************/

/* CB XX ߂̃e[uWvłB
 * [note]
 *	* ED XX ߂̈́A0oCgłB
 */
#ifndef Z80_ASM	/************************************************************/

Z80FN_("CB", CB)
{
	const Z80OP* op;

	FETCH(code);
	op = &z80_op_table_CB[code];
	ASSERT(op->argc == 0);
	EXEC();
}

#else	/********************************************************************/

asm("
	.code
	.align 1
	.global z80op_CB
z80op_CB:
	pushn %r2
	;
	ld.w %r0, %r12			; z80
	xld.uh %r1, [%r0+22]		; PC
	;
	xld.w %r11, [%r0+48]		; code = z80->read(z80, PC++)
	ld.w %r12, %r0
	ld.uh %r13, %r1
	call.d %r11
	add %r1, 1			; *delay*
	ld.ub %r13, %r10
	;
	xld.w %r2, z80_op_table_CB	; op = &z80_op_table_CB[code]
	xsll %r10, 3
	add %r2, %r10
	;
	xld.w %r10, [%r2+0]		; op->proc(z80, code, arg1, arg2)
	xld.h [%r0+22], %r1		; PC *anti-interlock*
	call.d %r10
	ld.w %r12, %r0			; *delay*
	;
	xld.w %r10, [%r0+36]		; CYCLE -= op->cycle
	xld.ub %r11, [%r2+5]
	sub %r10, %r11
	xld.w [%r0+36], %r10
	;
	popn %r2
	ret
");

#endif	/********************************************************************/

/* ED XX ߂̃e[uWvłB
 * [note]
 *	* ED XX ߂̈́A0oCg܂2oCgłB
 */
#ifndef Z80_ASM	/************************************************************/

Z80FN_("ED", ED)
{
	const Z80OP* op;

	FETCH(code);
	op = &z80_op_table_ED[code];
	ASSERT(op->argc == 0 || op->argc == 2);
	if(op->argc) {
		FETCH(arg1);
		FETCH(arg2);
	}
	EXEC();
}

#else	/********************************************************************/

asm("
	.code
	.align 1
	.global z80op_ED
z80op_ED:
	pushn %r2
	xsub %sp, %sp, 4
	;
	ld.w %r0, %r12			; z80
	xld.uh %r1, [%r0+22]		; PC
	;
	xld.w %r11, [%r0+48]		; code = z80->read(z80, PC++)
	ld.w %r12, %r0
	ld.uh %r13, %r1
	call.d %r11
	add %r1, 1			; *delay*
	ld.ub %r13, %r10
	;
	xld.w %r2, z80_op_table_ED	; op = &z80_op_table_ED[code]
	xsll %r10, 3
	add %r2, %r10
	;
	xld.ub %r10, [%r2+4]		; argc == 0 ?
	xcmp %r10, 0	
	xjreq z80op_ED_L10
	xld.b [%sp+0], %r13		; (save code)
	;
	xld.w %r11, [%r0+48]		; arg1 = z80->read(z80, PC++)
	ld.w %r12, %r0
	ld.uh %r13, %r1
	call.d %r11
	add %r1, 1			; *delay*
	xld.b [%sp+1], %r10		; (save arg1)
	;
	xld.w %r11, [%r0+48]		; arg2 = z80->read(z80, PC++)
	ld.w %r12, %r0
	ld.w %r13, %r1
	call.d %r11
	add %r1, 1			; *delay*
	ld.ub %r15, %r10
	;
	xld.ub %r14, [%sp+1]		; (restore arg1)
	xld.ub %r13, [%sp+0]		; (restore code)
z80op_ED_L10:
	;
	xld.w %r10, [%r2+0]		; op->proc(z80, code, arg1, arg2)
	xld.h [%r0+22], %r1		; PC *anti-interlock*
	call.d %r10
	ld.w %r12, %r0			; *delay*
	;
	xld.w %r10, [%r0+36]		; CYCLE -= op->cycle
	xld.ub %r11, [%r2+5]
	sub %r10, %r11
	xld.w [%r0+36], %r10
	;
	xadd %sp, %sp, 4
	popn %r2
	ret
");

#endif	/********************************************************************/

/* DD XX ߂̃e[uWvłB
 */
#ifndef Z80_ASM	/************************************************************/

Z80FN_("DD", DD)
{
	const Z80OP* op;

	FETCH(code);
	op = &z80_op_table_DD[code];
	if(op->argc >= 1) {
		FETCH(arg1);
		if(op->argc >= 2) {
			FETCH(arg2);
		}
	}
	EXEC();
}

#else	/********************************************************************/

asm("
	.code
	.align 1
	.global z80op_DD
z80op_DD:
	pushn %r2
	xsub %sp, %sp, 4
	;
	ld.w %r0, %r12			; z80
	xld.uh %r1, [%r0+22]		; PC
	;
	xld.w %r11, [%r0+48]		; code = z80->read(z80, PC++)
	ld.w %r12, %r0
	ld.uh %r13, %r1
	call.d %r11
	add %r1, 1			; *delay*
	ld.ub %r13, %r10
	;
	xld.w %r2, z80_op_table_DD	; op = &z80_op_table_DD[code]
	xsll %r10, 3
	add %r2, %r10
	;
	xld.ub %r10, [%r2+4]		; argc == 0 ?
	xcmp %r10, 0	
	xjreq z80op_DD_L10
	xld.b [%sp+0], %r13		; (save code)
	;
	xld.w %r11, [%r0+48]		; arg1 = z80->read(z80, PC++)
	ld.w %r12, %r0
	ld.uh %r13, %r1
	call.d %r11
	add %r1, 1			; *delay*
	ld.ub %r14, %r10
	;
	xld.ub %r10, [%r2+4]		; argc == 1 ?
	xcmp %r10, 1	
	xjreq z80op_DD_L20
	xld.b [%sp+1], %r14		; (save arg1)
	;
	xld.w %r11, [%r0+48]		; arg2 = z80->read(z80, PC++)
	ld.w %r12, %r0
	ld.w %r13, %r1
	call.d %r11
	add %r1, 1			; *delay*
	ld.ub %r15, %r10
	;
	xld.ub %r14, [%sp+1]		; (restore arg1)
z80op_DD_L20:
	xld.ub %r13, [%sp+0]		; (restore code)
z80op_DD_L10:
	;
	xld.w %r10, [%r2+0]		; op->proc(z80, code, arg1, arg2)
	xld.h [%r0+22], %r1		; PC *anti-interlock*
	call.d %r10
	ld.w %r12, %r0			; *delay*
	;
	xld.w %r10, [%r0+36]		; CYCLE -= op->cycle
	xld.ub %r11, [%r2+5]
	sub %r10, %r11
	xld.w [%r0+36], %r10
	;
	xadd %sp, %sp, 4
	popn %r2
	ret
");

#endif	/********************************************************************/

/* FD XX ߂̃e[uWvłB
 */
#ifndef Z80_ASM	/************************************************************/

Z80FN_("FD", FD)
{
	const Z80OP* op;

	FETCH(code);
	op = &z80_op_table_FD[code];
	if(op->argc >= 1) {
		FETCH(arg1);
		if(op->argc >= 2) {
			FETCH(arg2);
		}
	}
	EXEC();
}

#else	/********************************************************************/

asm("
	.code
	.align 1
	.global z80op_FD
z80op_FD:
	pushn %r2
	xsub %sp, %sp, 4
	;
	ld.w %r0, %r12			; z80
	xld.uh %r1, [%r0+22]		; PC
	;
	xld.w %r11, [%r0+48]		; code = z80->read(z80, PC++)
	ld.w %r12, %r0
	ld.uh %r13, %r1
	call.d %r11
	add %r1, 1			; *delay*
	ld.ub %r13, %r10
	;
	xld.w %r2, z80_op_table_FD	; op = &z80_op_table_FD[code]
	xsll %r10, 3
	add %r2, %r10
	;
	xld.ub %r10, [%r2+4]		; argc == 0 ?
	xcmp %r10, 0	
	xjreq z80op_FD_L10
	xld.b [%sp+0], %r13		; (save code)
	;
	xld.w %r11, [%r0+48]		; arg1 = z80->read(z80, PC++)
	ld.w %r12, %r0
	ld.uh %r13, %r1
	call.d %r11
	add %r1, 1			; *delay*
	ld.ub %r14, %r10
	;
	xld.ub %r10, [%r2+4]		; argc == 1 ?
	xcmp %r10, 1	
	xjreq z80op_FD_L20
	xld.b [%sp+1], %r14		; (save arg1)
	;
	xld.w %r11, [%r0+48]		; arg2 = z80->read(z80, PC++)
	ld.w %r12, %r0
	ld.w %r13, %r1
	call.d %r11
	add %r1, 1			; *delay*
	ld.ub %r15, %r10
	;
	xld.ub %r14, [%sp+1]		; (restore arg1)
z80op_FD_L20:
	xld.ub %r13, [%sp+0]		; (restore code)
z80op_FD_L10:
	;
	xld.w %r10, [%r2+0]		; op->proc(z80, code, arg1, arg2)
	xld.h [%r0+22], %r1		; PC *anti-interlock*
	call.d %r10
	ld.w %r12, %r0			; *delay*
	;
	xld.w %r10, [%r0+36]		; CYCLE -= op->cycle
	xld.ub %r11, [%r2+5]
	sub %r10, %r11
	xld.w [%r0+36], %r10
	;
	xadd %sp, %sp, 4
	popn %r2
	ret
");

#endif	/********************************************************************/

/* DD CB d XX ߂̃e[uWvłB
 * [note]
 *	* DD CB d XX ߂̈1oCgłB
 *	  uAIyR[hv̏ł邱ƂɒӂĂB
 */
#ifndef Z80_ASM	/************************************************************/

Z80FN_("DDCB", DDCB)
{
	const Z80OP* op;

	FETCH(arg1);
	FETCH(code);
	op = &z80_op_table_DDCB[code];
	ASSERT(op->argc == 1);
	EXEC();
}

#else	/********************************************************************/

asm("
	.code
	.align 1
	.global z80op_DDCB
z80op_DDCB:
	pushn %r2
	xsub %sp, %sp, 4
	;
	ld.w %r0, %r12			; z80
	xld.uh %r1, [%r0+22]		; PC
	;
	xld.w %r11, [%r0+48]		; arg1 = z80->read(z80, PC++)
	ld.w %r12, %r0
	ld.uh %r13, %r1
	call.d %r11
	add %r1, 1			; *delay*
	xld.b [%sp+0], %r10		; (save arg1)
	;
	xld.w %r11, [%r0+48]		; code = z80->read(z80, PC++)
	ld.w %r12, %r0
	ld.uh %r13, %r1
	call.d %r11
	add %r1, 1			; *delay*
	ld.ub %r13, %r10
	;
	xld.w %r2, z80_op_table_DDCB	; op = &z80_op_table_DDCB[code]
	xsll %r10, 3
	add %r2, %r10
	;
	xld.ub %r14, [%sp+0]		; (restore arg1)
	;
	xld.w %r10, [%r2+0]		; op->proc(z80, code, arg1, arg2)
	xld.h [%r0+22], %r1		; PC *anti-interlock*
	call.d %r10
	ld.w %r12, %r0			; *delay*
	;
	xld.w %r10, [%r0+36]		; CYCLE -= op->cycle
	xld.ub %r11, [%r2+5]
	sub %r10, %r11
	xld.w [%r0+36], %r10
	;
	xadd %sp, %sp, 4
	popn %r2
	ret
");

#endif	/********************************************************************/

/* FD CB d XX ߂̃e[uWvłB
 * [note]
 *	* FD CB d XX ߂̈1oCgłB
 *	  uAIyR[hv̏ł邱ƂɒӂĂB
 */
#ifndef Z80_ASM	/************************************************************/

Z80FN_("FDCB", FDCB)
{
	const Z80OP* op;

	FETCH(arg1);
	FETCH(code);
	op = &z80_op_table_FDCB[code];
	ASSERT(op->argc == 1);
	EXEC();
}

#else	/********************************************************************/

asm("
	.code
	.align 1
	.global z80op_FDCB
z80op_FDCB:
	pushn %r2
	xsub %sp, %sp, 4
	;
	ld.w %r0, %r12			; z80
	xld.uh %r1, [%r0+22]		; PC
	;
	xld.w %r11, [%r0+48]		; arg1 = z80->read(z80, PC++)
	ld.w %r12, %r0
	ld.uh %r13, %r1
	call.d %r11
	add %r1, 1			; *delay*
	xld.b [%sp+0], %r10		; (save arg1)
	;
	xld.w %r11, [%r0+48]		; code = z80->read(z80, PC++)
	ld.w %r12, %r0
	ld.uh %r13, %r1
	call.d %r11
	add %r1, 1			; *delay*
	ld.ub %r13, %r10
	;
	xld.w %r2, z80_op_table_FDCB	; op = &z80_op_table_FDCB[code]
	xsll %r10, 3
	add %r2, %r10
	;
	xld.ub %r14, [%sp+0]		; (restore arg1)
	;
	xld.w %r10, [%r2+0]		; op->proc(z80, code, arg1, arg2)
	xld.h [%r0+22], %r1		; PC *anti-interlock*
	call.d %r10
	ld.w %r12, %r0			; *delay*
	;
	xld.w %r10, [%r0+36]		; CYCLE -= op->cycle
	xld.ub %r11, [%r2+5]
	sub %r10, %r11
	xld.w [%r0+36], %r10
	;
	xadd %sp, %sp, 4
	popn %r2
	ret
");

#endif	/********************************************************************/

/****************************************************************************
 *
 ****************************************************************************/

/* peBtOZô߂̕⏕[`łB
 * [in]
 *	n		ׂlB
 * [out]
 *	߂l		1̃rbgȂ4Ԃ܂B
 *			1̃rbgȂ0Ԃ܂B
 * [note]
 *	* Z80̃peBtO͋peBłB
 *	* peBtÖʒubit2Ȃ̂ŁA4or0Ԃ܂B
 */
#ifndef Z80_ASM	/************************************************************/

unsigned char
z80_PARITY(unsigned char n)
{
	int p = 4;
	while(n) {
		if(n & 1) p ^= 4;
		n >>= 1;
	}
	return p;
}

#else	/********************************************************************/

asm("
	.code
	.align 1
	.global z80_PARITY
z80_PARITY:
	xrr %r12, 8			; ׂl8rbgɈړB
	xjreq.d z80_PARITY_L20		; 1̃rbgȂ΁A[vɓ܂B
	ld.w %r10, 4			; peBȂ̂ŁAԂONB*delay*
z80_PARITY_L10:
	scan1 %r11, %r12		; 1̃rbgʒuB(K݂)
	sll %r12, %r11			; rbgŏʂփVtgB
	sll %r12, 1			; rbg֎̂ĂB
	xjrne.d z80_PARITY_L10		; ܂1̃rbgcĂA[vB
	xor %r10, 4			; rbg̃peB]B*delay*
z80_PARITY_L20:
	ret
");

#endif	/********************************************************************/

/* AZu[`g߂́AC33=>Z80tOϊ\B
 */
#ifdef Z80_ASM
asm("
	.code			; C33(C,V,Z,N) => Z80(S,Z,V,C)
	.global z80_FLAG_CONV	; ====C33====     ====Z80====
z80_FLAG_CONV:			; ----CVZN        SZ---V-C   
	.byte	0x00		; 00000000 00     00000000 00
	.byte	0x80		; 00000001 01     10000000 80
	.byte	0x40		; 00000010 02     01000000 40
	.byte	0xc0		; 00000011 03     11000000 c0
	.byte	0x04		; 00000100 04     00000100 04
	.byte	0x84		; 00000101 05     10000100 84
	.byte	0x44		; 00000110 06     01000100 44
	.byte	0xc4		; 00000111 07     11000100 c4
	.byte	0x01		; 00001000 08     00000001 01
	.byte	0x81		; 00001001 09     10000001 81
	.byte	0x41		; 00001010 0a     01000001 41
	.byte	0xc1		; 00001011 0b     11000001 c1
	.byte	0x05		; 00001100 0c     00000101 05
	.byte	0x85		; 00001101 0d     10000101 85
	.byte	0x45		; 00001110 0e     01000101 45
	.byte	0xc5		; 00001111 0f     11000101 c5
");
#endif /*Z80_ASM*/

/****************************************************************************
 *	General-Purpose Arithmetic and CPU Control Group
 ****************************************************************************/

/* 8rbgZs܂B
 * [in/out]
 *	A <= A + n
 * [flags]
 *	S	0/1	Doc'd
 *	Z	0/1	Doc'd
 *	Y	0/1	UnDoc	ʂbit5Rs[܂B
 *	H	0/1	Doc'd
 *	X	0/1	UnDoc	ʂbit3Rs[܂B
 *	V	0/1	Doc'd
 *	N	0  	Doc'd
 *	C	0/1	Doc'd
 */
#ifndef Z80_ASM	/************************************************************/

void
z80_ADD8(Z80* z80, unsigned char n)
{
	int tmp = A + n;
	F = (tmp & 0xa8) |				/* S,Y,X */
	    (!(tmp & 0xff) << 6) |			/* Z */
	    (((A & 0xf) + (n & 0xf)) & 0x10) |		/* H */
	    ((~(A ^ n) & (A ^ tmp) & 0x80) >> 5) |	/* V */
							/* N */
	    ((tmp >> 8) & 1);				/* C */
	A = tmp;
}

#else	/********************************************************************/

asm("
	.code
	.align 1
	.global z80_ADD8
z80_ADD8:
	xld.ub %r14, [%r12+1]			; A
	ld.w %r11, %r13				; *anti-interlock*
	ld.w %r10, %r14
	xrr %r11, 8
	xrr %r10, 8
	add %r10, %r11
	;
	ld.w %r11, %psr				; F(S,Z,V,C)
	xand %r11, %r11, 0xf
	xld.ub %r11, [%r11+z80_FLAG_CONV]
	;
	xrl %r10, 8
	xld.b [%r12+1], %r10			; A
	;
	xand %r10, %r10, 0x28			; F(Y,X)
	or %r11, %r10
	;
	xand %r14, %r14, 0xf			; F(H)
	xand %r13, %r13, 0xf
	add %r14, %r13
	xand %r14, %r14, 0x10
	or %r11, %r14
	;
	xld.b [%r12+0], %r11			; F
	;
	ret
");

#endif	/********************************************************************/

/* 8rbgZ(L[)s܂B
 * [in/out]
 *	A <= A + n + F(C)
 * [flags]
 *	S	0/1	Doc'd
 *	Z	0/1	Doc'd
 *	Y	0/1	UnDoc	ʂbit5Rs[܂B
 *	H	0/1	Doc'd
 *	X	0/1	UnDoc	ʂbit3Rs[܂B
 *	V	0/1	Doc'd
 *	N	0  	Doc'd
 *	C	0/1	Doc'd
 */
void
z80_ADC8(Z80* z80, unsigned char n)
{
	int c = F & 1;
	int tmp = A + n + c;
	F = (tmp & 0xa8) |				/* S,Y,X */
	    (!(tmp & 0xff) << 6) |			/* Z */
	    (((A & 0xf) + (n & 0xf) + c) & 0x10) |	/* H */
	    ((~(A ^ n) & (A ^ tmp) & 0x80) >> 5) |	/* V */
							/* N */
	    ((tmp >> 8) & 1);				/* C */
	A = tmp;
}

/* 8rbgZs܂B
 * [in/out]
 *	A <= A - n
 * [flags]
 *	S	0/1	Doc'd
 *	Z	0/1	Doc'd
 *	Y	0/1	UnDoc	ʂbit5Rs[܂B
 *	H	0/1	Doc'd
 *	X	0/1	UnDoc	ʂbit3Rs[܂B
 *	V	0/1	Doc'd
 *	N	  1	Doc'd
 *	C	0/1	Doc'd
 */
#ifndef Z80_ASM	/************************************************************/

void
z80_SUB8(Z80* z80, unsigned char n)
{
	int tmp = A - n;
	F = (tmp & 0xa8) |				/* S,Y,X */
	    (!(tmp & 0xff) << 6) |			/* Z */
	    (((A & 0xf) - (n & 0xf)) & 0x10) |		/* H */
	    (((A ^ n) & (A ^ tmp) & 0x80) >> 5) |	/* V */
	    2 |						/* N */
	    ((tmp >> 8) & 1);				/* C */
	A = tmp;
}


#else	/********************************************************************/

asm("
	.code
	.align 1
	.global z80_SUB8
z80_SUB8:
	xld.ub %r14, [%r12+1]			; A
	ld.w %r11, %r13				; *anti-interlock*
	ld.w %r10, %r14
	xrr %r11, 8
	xrr %r10, 8
	sub %r10, %r11
	;
	ld.w %r11, %psr				; F(S,Z,V,C)
	xand %r11, %r11, 0xf
	xld.ub %r11, [%r11+z80_FLAG_CONV]
	;
	xrl %r10, 8
	xld.b [%r12+1], %r10			; A
	;
	xand %r10, %r10, 0x28			; F(Y,X)
	or %r11, %r10
	;
	xand %r14, %r14, 0xf			; F(H)
	xand %r13, %r13, 0xf
	sub %r14, %r13
	xand %r14, %r14, 0x10
	or %r11, %r14
	;
	xor %r11, 2				; F(N)
	;
	xld.b [%r12+0], %r11			; F
	;
	ret
");

#endif	/********************************************************************/

/* 8rbgZ({[)s܂B
 * [in/out]
 *	A <= A - n - F(C)
 * [flags]
 *	S	0/1	Doc'd
 *	Z	0/1	Doc'd
 *	Y	0/1	UnDoc	ʂbit5Rs[܂B
 *	H	0/1	Doc'd
 *	X	0/1	UnDoc	ʂbit3Rs[܂B
 *	V	0/1	Doc'd
 *	N	  1	Doc'd
 *	C	0/1	Doc'd
 */
void
z80_SBC8(Z80* z80, unsigned char n)
{
	int c = F & 1;
	int tmp = A - n - c;
	F = (tmp & 0xa8) |				/* S,Y,X */
	    (!(tmp & 0xff) << 6) |			/* Z */
	    (((A & 0xf) - (n & 0xf) - c) & 0x10) |	/* H */
	    (((A ^ n) & (A ^ tmp) & 0x80) >> 5) |	/* V */
	    2 |						/* N */
	    ((tmp >> 8) & 1);				/* C */
	A = tmp;
}

/* _ς߂܂B
 * [in/out]
 *	A <= A & n
 * [flags]
 *	S	0/1	Doc'd
 *	Z	0/1	Doc'd
 *	Y	0/1	UnDoc	ʂbit5Rs[܂B
 *	H	  1	Doc'd
 *	X	0/1	UnDoc	ʂbit3Rs[܂B
 *	P	0/1	Doc'd
 *	N	0  	Doc'd
 *	C	0  	Doc'd
 */
void
z80_AND(Z80* z80, unsigned char n)
{
	A &= n;
	F = (A & 0xa8) |	/* S,Y,X */
	    (!A << 6) |		/* Z */
	    0x10 |		/* H */
	    PARITY(A);		/* P */
				/* N */
				/* C */
}

/* _a߂܂B
 * [in/out]
 *	A <= A & n
 * [flags]
 *	S	0/1	Doc'd
 *	Z	0/1	Doc'd
 *	Y	0/1	UnDoc	ʂbit5Rs[܂B
 *	H	0  	Doc'd
 *	X	0/1	UnDoc	ʂbit3Rs[܂B
 *	P	0/1	Doc'd
 *	N	0  	Doc'd
 *	C	0  	Doc'd
 */
void
z80_OR(Z80* z80, unsigned char n)
{
	A |= n;
	F = (A & 0xa8) |	/* S,Y,X */
	    (!A << 6) |		/* Z */
				/* H */
	    PARITY(A);		/* P */
				/* N */
				/* C */
}

/* r_a߂܂B
 * [in/out]
 *	A <= A ^ n
 * [flags]
 *	S	0/1	Doc'd
 *	Z	0/1	Doc'd
 *	Y	0/1	UnDoc	ʂbit5Rs[܂B
 *	H	0  	Doc'd
 *	X	0/1	UnDoc	ʂbit3Rs[܂B
 *	P	0/1	Doc'd
 *	N	0  	Doc'd
 *	C	0  	Doc'd
 */
void
z80_XOR(Z80* z80, unsigned char n)
{
	A ^= n;
	F = (A & 0xa8) |	/* S,Y,X */
	    (!A << 6) |		/* Z */
				/* H */
	    PARITY(A);		/* P */
				/* N */
				/* C */
}

/* rs܂B
 * [in/out]
 *	̂Ă <= A - n
 * [flags]
 *	S	0/1	Doc'd
 *	Z	0/1	Doc'd
 *	Y	0/1	UnDoc	rlbit5Rs[܂B
 *	H	0/1	Doc'd
 *	X	0/1	UnDoc	rlbit3Rs[܂B
 *	V	0/1	Doc'd
 *	N	  1	Doc'd
 *	C	0/1	Doc'd
 * [note]
 *	* tOWX^bit5,3́Aʂł͂ȂrΏےl𔽉f܂B
 *	  ̖߂AgptOrbg̐U镑łBv!!
 */
void
z80_CP(Z80* z80, unsigned char n)
{
	int tmp = A - n;
	F = (tmp & 0x80) |				/* S */
	    (n & 0x28) |				/* Y,X */
	    (!(tmp & 0xff) << 6) |			/* Z */
	    (((A & 0xf) - (n & 0xf)) & 0x10) |		/* H */
	    (((A ^ n) & (A ^ tmp) & 0x80) >> 5) |	/* V */
	    2 |						/* N */
	    ((tmp >> 8) & 1);				/* C */
}

/* 8rbglCNg܂B
 * [in/out]
 *	߂l <= r + 1
 * [flags]
 *	S	0/1	Doc'd
 *	Z	0/1	Doc'd
 *	Y	0/1	UnDoc	ʂbit5Rs[܂B
 *	H	0/1	Doc'd
 *	X	0/1	UnDoc	ʂbit3Rs[܂B
 *	P	0/1	Doc'd
 *	N	0  	Doc'd
 *	C	---	Doc'd
 */
unsigned char
z80_INC8(Z80* z80, unsigned char r)
{
	r++;
	F = (r & 0xa8) |		/* S,Y,X */
	    (!r << 6) |			/* Z */
	    (!(r & 0xf) << 4) |		/* H */
	    ((r == 0x80) << 2) |	/* V */
					/* N */
	    (F & 1);			/* C */
	return r;
}

/* 8rbglfNg܂B
 * [in/out]
 *	߂l <= r + 1
 * [flags]
 *	S	0/1	Doc'd
 *	Z	0/1	Doc'd
 *	Y	0/1	UnDoc	ʂbit5Rs[܂B
 *	H	0/1	Doc'd
 *	X	0/1	UnDoc	ʂbit3Rs[܂B
 *	P	0/1	Doc'd
 *	N	  1	Doc'd
 *	C	---	Doc'd
 */
unsigned char
z80_DEC8(Z80* z80, unsigned char r)
{
	r--;
	F = (r & 0xa8) |		/* S,Y,X */
	    (!r << 6) |			/* Z */
	    (((r & 0xf) == 0xf) << 4) |	/* H */
	    ((r == 0x7f) << 2) |	/* V */
	    2 |				/* N */
	    (F & 1);			/* C */
	return r;
}

/****************************************************************************
 *	16-Bit Arithmetic Group
 ****************************************************************************/

/* 16rbgZs܂B
 * [in/out]
 *	߂l <= rr + nn
 * [flags]
 *	S	---	Doc'd
 *	Z	---	Doc'd
 *	Y	0/1	UnDoc	ʂ̏ʃoCgbit5Rs[܂B
 *	H	0/1	UnDoc	ʂ̏ʃoCg̃n[tL[łB
 *	X	0/1	UnDoc	ʂ̏ʃoCgbit3Rs[܂B
 *	V	---	Doc'd
 *	N	  0	Doc'd
 *	C	0/1	Doc'd
 */
unsigned short
z80_ADD16(Z80* z80, unsigned short rr, unsigned short nn)
{
	int tmp = rr + nn;
	F = (F & 0xc4) |					/* S,Z,V */
	    ((tmp >> 8) & 0x28) |				/* Y,X */
	    ((((rr & 0xfff) + (nn & 0xfff)) >> 8) & 0x10) |	/* H */
								/* N */
	    ((tmp >> 16) & 1);					/* C */
	return tmp;
}

/* 16rbgZ(L[)s܂B
 * [in/out]
 *	߂l <= rr + nn + F(C)
 * [flags]
 *	S	0/1	Doc'd
 *	Z	0/1	Doc'd
 *	Y	0/1	UnDoc	ʂ̏ʃoCgbit5Rs[܂B
 *	H	0/1	UnDoc	ʂ̏ʃoCg̃n[tL[łB
 *	X	0/1	UnDoc	ʂ̏ʃoCgbit3Rs[܂B
 *	V	0/1	Doc'd
 *	N	  0	Doc'd
 *	C	0/1	Doc'd
 */
unsigned short
z80_ADC16(Z80* z80, unsigned short rr, unsigned short nn)
{
	int c = F & 1;
	int tmp = rr + nn + c;
	F = ((tmp >> 8) & 0xa8) |				/* S,Y,X */
	    (!(tmp & 0xffff) << 6) |				/* Z */
	    ((((rr & 0xfff) + (nn & 0xfff) + c) >> 8) & 0x10) |	/* H */
	    ((~(rr ^ nn) & (rr ^ tmp) & 0x8000) >> 13) |	/* V */
								/* N */
	    ((tmp >> 16) & 1);					/* C */
	return tmp;
}

/* 16rbgZ({[)s܂B
 * [in/out]
 *	߂l <= rr - nn - F(C)
 * [flags]
 *	S	0/1	Doc'd
 *	Z	0/1	Doc'd
 *	Y	0/1	UnDoc	ʂ̏ʃoCgbit5Rs[܂B
 *	H	0/1	UnDoc	ʂ̏ʃoCg̃n[tL[łB
 *	X	0/1	UnDoc	ʂ̏ʃoCgbit3Rs[܂B
 *	V	0/1	Doc'd
 *	N	1  	Doc'd
 *	C	0/1	Doc'd
 */
unsigned short
z80_SBC16(Z80* z80, unsigned short rr, unsigned short nn)
{
	int c = F & 1;
	int tmp = rr - nn - c;
	F = ((tmp >> 8) & 0xa8) |				/* S,Y,X */
	    (!(tmp & 0xffff) << 6) |				/* Z */
	    ((((rr & 0xfff) - (nn & 0xfff) - c) >> 8) & 0x10) |	/* H */
	    (((rr ^ nn) & (rr ^ tmp) & 0x8000) >> 13) |		/* V */
	    2 |							/* N */
	    ((tmp >> 16) & 1);					/* C */
	return tmp;
}

/****************************************************************************
 *	Rotate and Shift Group
 ****************************************************************************/

/* [e[g(L[)s܂B
 * [in/out]
 *	߂l <= RL r
 * [flags]
 *	S	0/1	Doc'd
 *	Z	0/1	Doc'd
 *	Y	0/1	UnDoc	ʂbit5Rs[܂B
 *	H	0  	Doc'd
 *	X	0/1	UnDoc	ʂbit3Rs[܂B
 *	P	0/1	Doc'd
 *	N	0  	Doc'd
 *	C	0/1	Doc'd
 */
unsigned char
z80_RL(Z80* z80, unsigned char r)
{
	unsigned char tmp = (r << 1) | (F & 1);
	F = (tmp & 0xa8) |	/* S,Y,X */
	    (!tmp << 6) |	/* Z */
				/* H */
	    PARITY(tmp) |	/* P */
				/* N */
	    (r >> 7);		/* C */
	return tmp;
}

/* [e[g(T[L)s܂B
 * [in/out]
 *	߂l <= RLC r
 * [flags]
 *	S	0/1	Doc'd
 *	Z	0/1	Doc'd
 *	Y	0/1	UnDoc	ʂbit5Rs[܂B
 *	H	0  	Doc'd
 *	X	0/1	UnDoc	ʂbit3Rs[܂B
 *	P	0/1	Doc'd
 *	N	0  	Doc'd
 *	C	0/1	Doc'd
 */
unsigned char
z80_RLC(Z80* z80, unsigned char r)
{
	unsigned char tmp = (r << 1) | (r >> 7);
	F = (tmp & 0xa8) |	/* S,Y,X */
	    (!tmp << 6) |	/* Z */
				/* H */
	    PARITY(tmp) |	/* P */
				/* N */
	    (r >> 7);		/* C */
	return tmp;
}

/* E[e[g(L[)s܂B
 * [in/out]
 *	߂l <= RR r
 * [flags]
 *	S	0/1	Doc'd
 *	Z	0/1	Doc'd
 *	Y	0/1	UnDoc	ʂbit5Rs[܂B
 *	H	0  	Doc'd
 *	X	0/1	UnDoc	ʂbit3Rs[܂B
 *	P	0/1	Doc'd
 *	N	0  	Doc'd
 *	C	0/1	Doc'd
 */
unsigned char
z80_RR(Z80* z80, unsigned char r)
{
	unsigned char tmp = (r >> 1) | (F << 7);
	F = (tmp & 0xa8) |	/* S,Y,X */
	    (!tmp << 6) |	/* Z */
				/* H */
	    PARITY(tmp) |	/* P */
				/* N */
	    (r & 1);		/* C */
	return tmp;
}

/* E[e[g(T[L)s܂B
 * [in/out]
 *	߂l <= RRC r
 * [flags]
 *	S	0/1	Doc'd
 *	Z	0/1	Doc'd
 *	Y	0/1	UnDoc	ʂbit5Rs[܂B
 *	H	0  	Doc'd
 *	X	0/1	UnDoc	ʂbit3Rs[܂B
 *	P	0/1	Doc'd
 *	N	0  	Doc'd
 *	C	0/1	Doc'd
 */
unsigned char
z80_RRC(Z80* z80, unsigned char r)
{
	unsigned char tmp = (r >> 1) | (r << 7);
	F = (tmp & 0xa8) |	/* S,Y,X */
	    (!tmp << 6) |	/* Z */
				/* H */
	    PARITY(tmp) |	/* P */
				/* N */
	    (r & 1);		/* C */
	return tmp;
}

/* ZpVtgs܂B
 * [in/out]
 *	߂l <= (char)r << 1
 * [flags]
 *	S	0/1	Doc'd
 *	Z	0/1	Doc'd
 *	Y	0/1	UnDoc	ʂbit5Rs[܂B
 *	H	0  	Doc'd
 *	X	0/1	UnDoc	ʂbit3Rs[܂B
 *	P	0/1	Doc'd
 *	N	0  	Doc'd
 *	C	0/1	Doc'd
 */
unsigned char
z80_SLA(Z80* z80, unsigned char r)
{
	unsigned char tmp = r << 1;
	F = (tmp & 0xa8) |	/* S,Y,X */
	    (!tmp << 6) |	/* Z */
				/* H */
	    PARITY(tmp) |	/* P */
				/* N */
	    (r >> 7);		/* C */
	return tmp;
}

/* EZpVtgs܂B
 * [in/out]
 *	߂l <= (char)r >> 1
 * [flags]
 *	S	0/1	Doc'd
 *	Z	0/1	Doc'd
 *	Y	0/1	UnDoc	ʂbit5Rs[܂B
 *	H	0  	Doc'd
 *	X	0/1	UnDoc	ʂbit3Rs[܂B
 *	P	0/1	Doc'd
 *	N	0  	Doc'd
 *	C	0/1	Doc'd
 */
unsigned char
z80_SRA(Z80* z80, unsigned char r)
{
	unsigned char tmp = (char)r >> 1;
	F = (tmp & 0xa8) |	/* S,Y,X */
	    (!tmp << 6) |	/* Z */
				/* H */
	    PARITY(tmp) |	/* P */
				/* N */
	    (r & 1);		/* C */
	return tmp;
}

/* E_Vtgs܂B
 * [in/out]
 *	߂l <= (unsigned char)r >> 1
 * [flags]
 *	S	0/1	Doc'd
 *	Z	0/1	Doc'd
 *	Y	0/1	UnDoc	ʂbit5Rs[܂B
 *	H	0  	Doc'd
 *	X	0/1	UnDoc	ʂbit3Rs[܂B
 *	P	0/1	Doc'd
 *	N	0  	Doc'd
 *	C	0/1	Doc'd
 */
unsigned char
z80_SRL(Z80* z80, unsigned char r)
{
	unsigned char tmp = r >> 1;
	F = (tmp & 0xa8) |	/* S,Y,X */
	    (!tmp << 6) |	/* Z */
				/* H */
	    PARITY(tmp) |	/* P */
				/* N */
	    (r & 1);		/* C */
	return tmp;
}

/****************************************************************************
 *	Bit Set, Reset, and Test Group
 ****************************************************************************/

/* rbgeXgs܂B
 * [in/out]
 *	̂Ă <= r & (1 << n)
 * [flags]
 *	S	0/1	UnDoc	ʂbit7Rs[܂B
 *	Z	0/1	Doc'd
 *	Y	???	UnDoc	GĂ悭킩܂c(^^;
 *	H	  1	Doc'd
 *	X	???	UnDoc	GĂ悭킩܂c(^^;
 *	P	0/1	UnDoc	ZtOƓlB(ʂ̋peBł)
 *	N	0  	Doc'd
 *	C	---	Doc'd
 */
void
z80_BIT(Z80* z80, unsigned char n, unsigned char r)
{
	unsigned char tmp = r & (1 << n);
	F = (tmp & 0x80) |		/* S */
	    (tmp ? 0x10 : 0x54) |	/* Z,H,P */
					/* N */
	    (F & 1);			/* C */
}

